// -*- C++ -*-


/// \addtogroup device
/// \{

#pragma once


#include "device.h"
#include "goopax_license.h"
#include "memory.h"

namespace goopax {

namespace impl {
namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(4.0.0) size_t goopax_4_num_devices(envmode, unsigned int, void (*)(), const char*, const uint8_t*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_get_devices(goopax_device*, envmode, unsigned int, void (*)(), const char*, const uint8_t*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) goopax_device_impl* goopax_4_get_default_device(envmode, unsigned int, void (*)(), const char*, const uint8_t*) noexcept;

GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_register_garbage(void*, void (*)(void*)) noexcept;
}
}

template<typename T>
static void garbage_deleter(void* p)
{
    delete static_cast<T*>(p);
}


static const uint8_t sig[256] = { GOOPAX_CERT_SIG };

inline vector<goopax_device> devices(envmode backend = env_GPU)
{
    /// \ingroup device
    /// \brief Get vector of available devices for the given backend(s)
    /// Devices that are both CUDA and OpenCL capable will show up both for the env_CUDA and for the env_CL backend,
    /// but they will only be returned once if the both env_CL and env_CUDA flags are selected. Equally, devices that support the Metal backend may also show up if the env_CL backend is selected without the env_METAL backend.
    /// \return vector of devices. If no usable devices are found, an empty list is returned.
    /// \param backend which backend(s) to use.
    vector<goopax_device> devices(interface::goopax_4_num_devices(backend, static_cast<unsigned int>(GOOPAX_VERSION_ID), goopax_debugfunc, GOOPAX_CERT, sig));
    interface::goopax_4_get_devices(devices.data(), backend, static_cast<unsigned int>(GOOPAX_VERSION_ID), goopax_debugfunc, GOOPAX_CERT, sig);
    return devices;
}

inline goopax_device default_device(envmode backend = env_GPU)
{
    /// \ingroup device
    /// \brief Get the fastest available device for the given backend(s)
    /// \return best suitable device. If no usable device is found, the returned value will be default-constructed, and goopax_device::valid() will return false on it.
    /// \param backend which backend(s) to use.
    return interface::goopax_4_get_default_device(backend, static_cast<unsigned int>(GOOPAX_VERSION_ID), goopax_debugfunc, GOOPAX_CERT, sig);
}

inline void wait_all_devices()
{
    check_errno(interface::goopax_4E_rd_wait_all(nullptr));
}
}

using impl::default_device;
using impl::devices;
using impl::goopax_device;
using impl::wait_all_devices;
}

/// \}
